"""
These types are used for tracking the package/modules/items annotated and exported by Rust code.
"""

from collections import defaultdict
from collections.abc import ItemsView, KeysView, MutableMapping, MutableSet, ValuesView
from dataclasses import dataclass, field
from pathlib import Path
from typing_extensions import Self
from typing import (
    Iterable,
    Iterator,
    TypeVar,
    overload,
)
import enum
import logging
import re

from .reader import Line

logger = logging.getLogger(__name__)

class PackageKind(enum.Enum):
    Annotation = "Annotation"
    Export = "Export"


class Kind(enum.Enum):
    """The sorts of things that can be annotated."""

    Struct = "struct"
    Enumeration = "enum"
    Function = "fn"
    # When discovered in a pymodule, we don't know the Rust type.
    Class = "class"
    Error = "error"

    def is_py_like(self, py: Self, /) -> bool:
        """Return ``True`` if the Python-argument is "like" this one; that is,::
        - they have the same kind, or
        - this is a struct or enum and the argument is a class or error.

        This is useful for comparing annotated Rust items (where we know the Rust item kind)
        to exported Python items (where we know the Python item kind).
        """
        match (self, py):
            case (Kind.Enumeration | Kind.Struct, Kind.Class | Kind.Error):
                return True
            case (x, y) if x is y:
                return True
        return False


class StubKind(enum.Enum):
    """A ``gen_stub_*`` attribute."""

    Function = "function"
    Methods = "methods"
    Class = "class"
    Enumeration = "class_enum"
    ComplexEnumeration = "class_complex_enum"

    def is_like(self, kind: Kind) -> bool:
        match (kind, self):
            case (Kind.Struct, StubKind.Class):
                return True
            case (
                Kind.Enumeration,
                (StubKind.Enumeration | StubKind.ComplexEnumeration),
            ):
                return True
            case (
                Kind.Class,
                (StubKind.Class | StubKind.Enumeration | StubKind.ComplexEnumeration),
            ):
                return True
            case (Kind.Function, StubKind.Function):
                return True
        return False


@dataclass
class StubAttr:
    """Represents information found with a ``pyo3_stub_gen`` annotation."""

    kind: StubKind
    module: str | None = None

    @classmethod
    def from_match(cls, stubgen_match: re.Match) -> Self:
        """Process #[gen_stub_...] annotations."""

        stub_kind = stubgen_match.group(1) or stubgen_match.group(3)
        if stub_kind is None:
            raise ValueError("Unable to determine stub kind")

        kind = StubKind(stub_kind)
        module: str | None = None

        if kind is StubKind.Function:
            props = PyO3Props.parse(stubgen_match.group(2) or stubgen_match.group(4))
            if not (module := props.gets("module")):
                logger.warning(f"Unknown module for stub {stubgen_match.group(0)}")

        return cls(kind, module)


class PyO3Props(dict[str, str | bool]):
    """Properties found in a PyO3 attribute, such as ``module = "name"`` or ``ord``.

    Use `parse` and `__str__` to convert to/from a string.
    You can check for the presence of a property with ordinary ``"prop" in props`` syntax,
    but you should prefer the type-specific versions which raise `TypeError`s
    if the property exists but doesn't match the expected type::

    - `gets(key)` returns a string-valued property
    - `is_(key)` checks for a boolean-valued property
    """

    def gets(self, key: str, default: str | None = None) -> str:
        if not (default is None or isinstance(default, str)):
            raise TypeError("if given, default must be str")
        if (value := super().get(key, default)) is not None and isinstance(value, str):
            return value
        if key in self:
            raise TypeError(f"expected {key} to be str, not {type(value)}")
        raise KeyError(key)

    def is_(self, key: str, default: bool = False) -> bool:
        if not (default is None or isinstance(default, bool)):
            raise TypeError("if given, default must be bool")
        if (value := super().get(key, default)) is not None and isinstance(value, bool):
            return value
        if key in self:
            raise TypeError(f"expected {key} to be bool, not {type(value)}")
        raise KeyError(key)

    @classmethod
    def parse(cls, prop_str: str) -> Self:
        """Convert props used in a ``#[whatever(arg1, arg2 = "val")]`` annotation.
        Note that the `prop_str` should just be the part between the parentheses.

        For the above example, the return value is ``{"arg1": True, "arg2": "val"}``.
        """

        result = cls()
        parts = [p.strip() for p in prop_str.split(",")]
        for part in parts:
            if "=" in part:
                key, val = map(str.strip, part.split("=", 1))
                if val.startswith('"') and val.endswith('"'):
                    val = val[1:-1]
                if key:
                    result[key] = val
            elif part:
                result[part] = True
        return result

    def __str__(self) -> str:
        """Convert back into a string."""
        return ", ".join(
            f'{k} = "{v}"' if isinstance(v, str) else str(k)
            for k, v in sorted(self.items())
        )

    def __sub__(self, keys: set[str]) -> Self:
        cls = self.__class__
        return cls((k, v) for k, v in self.items() if k not in keys)


@dataclass(frozen=True, order=True)
class Item:
    """Something that can be included in a Module (e.g., a class, function, exception, etc.)."""

    rust_name: str
    python_name: str
    kind: Kind = field(compare=False, hash=False)
    path: Path = field(compare=False, hash=False)
    line: Line = field(compare=False, hash=False)
    props: PyO3Props = field(compare=False, hash=False, default_factory=PyO3Props)
    stub_attr: StubAttr | None = field(compare=False, hash=False, default=None)

    def __str__(self) -> str:
        name = self.rust_name
        if self.rust_name != self.python_name:
            name = f"{name} (py: {self.python_name})"
        return f"{self.kind.value} '{name}' in {self.path}@{self.line.num}"

    def has_stub(self) -> bool:
        return self.stub_attr is not None


@dataclass
class Module(MutableSet[Item]):
    """A collection of `Item`s in the same Python module."""

    _items: set[Item] = field(default_factory=set)
    submodule: bool = True
    path: Path = Path(".")
    line_num: int = -1
    props: PyO3Props = field(default_factory=PyO3Props)
    _fixed_enums: set[str] = field(default_factory=set)

    def __contains__(self, key: object) -> bool:
        return key in self._items

    def __iter__(self) -> Iterator[Item]:
        return self._items.__iter__()

    def __len__(self) -> int:
        return self._items.__len__()

    def update(self, items: Iterable[Item]) -> None:
        return self._items.update(items)

    def discard(self, value: Item) -> None:
        return self._items.discard(value)

    def add(self, item: Item) -> None:
        logger.info(f"Adding {item}")
        self._items.add(item)

    def find_rust(self, item: Item | str) -> Item | None:
        """Return an `item` in this `Module` that has the same Rust name as the given `item`."""
        target = item.rust_name if isinstance(item, Item) else item
        return next((i for i in self if i.rust_name == target), None)

_T = TypeVar("_T")

@dataclass
class Package(MutableMapping[str, Module]):
    """A collection of `Module`s."""

    _modules: defaultdict[str, Module] = field(
        default_factory=lambda: defaultdict(Module)
    )

    # This tracks ``rust_type_name -> methods`` for `impl` blocks as we discover them,
    # for when we don't yet know which Python module the implemented type belongs to.
    _methods: Module = field(default_factory=Module)

    def find_rust(self, item: Item | str) -> tuple[str, Item] | None:
        """Return the name of a `Module` and an `Item` within it
        that has the same Rust name as the given `item`,
        but is distinct from the given `item.`

        If the given `item` is a `str` instead of an actual `Item`,
        the search is instead for an `Item` with that Rust name,
        and distinctness does not apply.

        The `Item`-argument form is useful for finding an `Item`
        that has been mistakenly placed in multiple modules,
        while `str`-argument form is useful for associating something with an `Item`
        when its module isn't immediately available (i.e., in an `impl` block).
        """
        return next(
            (
                (name, found)
                for name, module in self._modules.items()
                if (found := module.find_rust(item)) is not None
                and (isinstance(item, str) or id(found) != id(item))
            ),
            None,
        )

    def __getitem__(self, name: str, /) -> Module:
        if name not in self._modules:
            logger.info(f"Creating module '{name}'.")
        return self._modules[name]

    @overload  # type: ignore
    def get(self, key: str, default: None = None, /) -> Module | None: ...
    @overload
    def get(self, key: str, default: Module, /) -> Module: ...
    @overload
    def get(self, key: str, default: _T, /) -> Module | _T: ...

    def get(self, name: str, default: _T | None = None, /) -> Module | _T | None:
        return self._modules.get(name, default)

    def __setitem__(self, name: str, module: Module, /) -> None:
        logger.info(f"Setting module '{name}'.")
        self._modules[name] = module

    def __delitem__(self, name: str, /) -> None:
        del self._modules[name]

    def __contains__(self, name: object) -> bool:
        if not isinstance(name, str):
            raise TypeError(f"Package keys must be strings, not {type(name)}")
        return name in self._modules

    def __iter__(self) -> Iterator[str]:
        return self._modules.__iter__()

    def __len__(self) -> int:
        return self._modules.__len__()

    def keys(self) -> KeysView[str]:
        return self._modules.keys()

    def values(self) -> ValuesView[Module]:
        return self._modules.values()

    def items(self) -> ItemsView[str, Module]:
        return self._modules.items()




from pathlib import Path

from pyo3_linter import Item, Line

def test_item():
    """Verify that hash and equality checks work as expected.

    Namely, we want to compare items on their Python+Rust names,
    but ignore the rest of their metadata.
    """
    i1 = Item("A", "B", "struct", Path("C"), Line(1, "struct B();"))
    i2 = Item("A", "B", "struct", Path("c"), Line(2, "struct B();"))
    i3 = Item("A", "b", "struct", Path("C"), Line(1, "struct b();"))
    assert hash(i1) == hash(i2)
    assert hash(i1) != hash(i3)
    assert i1 in {i2}
    assert i2 in {i1}
    assert i1 not in {i3}
    assert i2 not in {i3}
    assert i3 not in {i1, i2}






from pathlib import Path

import pytest

def pytest_addoption(parser):
    parser.addoption("--generate", action="store_true", help="regenerate expected test data")

def pytest_generate_tests(metafunc: pytest.Metafunc):
    if "root" in metafunc.fixturenames:
        if (metafunc.module is None
                or not (data_dir := Path(metafunc.module.__file__).parent / "data").exists()
                or not data_dir.is_dir()):
            raise FileNotFoundError("unable to find test data directory")
        paths = [p for p in data_dir.iterdir() if p.is_dir()]
        metafunc.parametrize("root", paths, ids=[p.name for p in paths])
    if "generating" in metafunc.fixturenames:
        metafunc.parametrize("generating", (metafunc.config.getoption("generate", False),))




"""
This script is a lint helper for our PyO3 wrappers.

Given a starting directory, it recursively searches it for ``*.rs`` files,
and attempts to extract PyO3 annotations and exports from the source files.
Afterward, it may print some messages about potential mistakes.
Run the script with ``--help`` to see its options.
"""

from pathlib import Path
import argparse
import sys
import logging

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger()

from pyo3_linter import find_possible_mistakes, print_package_info, process_dir


def main():
    args = get_parser().parse_args()

    if args.log_level is not None:
        logger.setLevel(args.log_level)

    annotated, exported = process_dir(args.base)

    issues = find_possible_mistakes(annotated, exported)
    if args.show_mistakes:
        for issue in issues:
            print(issue.message)

    if args.show_package:
        print_package_info(annotated)

    if issues:
        print(f"\n {len(issues)} potential issue(s) discovered.", file=sys.stderr)
        if not args.show_mistakes:
            print("  (use --show-mistakes to see)", file=sys.stderr)
        sys.exit(1)


def get_parser() -> argparse.ArgumentParser:
    DEFAULT_PATH = Path("./src")

    parser = argparse.ArgumentParser(description="Lint quil-py source files.")

    parser.add_argument(
        "-b",
        "--base",
        metavar="PATH",
        type=Path,
        help=f"the base path to source files (default: '{DEFAULT_PATH}')",
        default=DEFAULT_PATH,
    )

    parser.add_argument(
        "-m",
        "--show-mistakes",
        action="store_true",
        default="",
        help="Show likely mistakes (default: enabled).",
    )

    parser.add_argument(
        "-M",
        "--no-show-mistakes",
        action="store_false",
        dest="show_mistakes",
        help="Don't show likely mistakes.",
    )

    parser.add_argument(
        "-p",
        "--show-package",
        action="store_true",
        help="Show package details (default: disabled).",
    )

    parser.add_argument(
        "-P",
        "--no-show-package",
        action="store_false",
        dest="show_package",
        default="",
        help="Don't show package details.",
    )

    parser.add_argument(
        "--log-level",
        help="set the logger level",
        choices=tuple(
            logging.getLevelName(level)
            for level in (logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR)
        ),
    )

    return parser


if __name__ == "__main__":
    main()



import inspect
import pickle
from typing import Optional

import pytest
from syrupy.assertion import SnapshotAssertion

from quil.instructions import Gate, Instruction, Jump, Qubit, QubitPlaceholder, Target, TargetPlaceholder
from quil.program import InstructionTarget, Program
from quil import QuilError


def test_pickle():
    program = Program.parse(
        """
DECLARE ro BIT[2]
H 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""
    )
    pickled = pickle.dumps(program)
    unpickled = pickle.loads(pickled)
    assert program == unpickled


def test_custom_resolver():
    qubit_placeholder = QubitPlaceholder()

    def qubit_resolver(qubit: QubitPlaceholder) -> Optional[int]:
        return {qubit_placeholder: 9}.get(qubit)

    target_placeholder = TargetPlaceholder("base-target")

    def target_resolver(target: TargetPlaceholder) -> Optional[str]:
        print("resolving target", target)
        return {target_placeholder: "test"}.get(target)

    program = Program()
    program.add_instructions(
        [
            Instruction.Gate(Gate("H", [], [Qubit.Placeholder(qubit_placeholder)], [])),
            Instruction.Jump(Jump(Target.Placeholder(target_placeholder))),
        ]
    )

    with pytest.raises(QuilError):
        program.to_quil()

    program.resolve_placeholders_with_custom_resolvers(target_resolver=target_resolver, qubit_resolver=qubit_resolver)

    print(program.to_quil_or_debug())

    assert program.to_quil() == "H 9\nJUMP @test\n"


def test_single_block_control_flow_analysis():
    program = Program.parse(
        """
DECLARE ro BIT[2]
H 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""
    )
    assert program.used_qubits == {Qubit.Fixed(0), Qubit.Fixed(1)}
    cfg = program.control_flow_graph()
    blocks = cfg.basic_blocks()
    assert len(blocks) == 1
    assert blocks[0].terminator is None
    block_program = Program()
    block_program.add_instructions(blocks[0].instructions)
    assert (
        block_program.to_quil()
        == """H 0
CNOT 0 1
MEASURE 0 ro[0]
MEASURE 1 ro[1]
"""
    )


def test_multi_block_control_flow_analysis():
    program = Program.parse(
        """
DECLARE ro BIT[1]
LABEL @start
MEASURE 0 ro[0]
JUMP-UNLESS @end ro[0]
X 0
JUMP @start
LABEL @end
HALT
"""
    )
    assert program.used_qubits == {Qubit.Fixed(0)}
    cfg = program.control_flow_graph()

    assert cfg.has_dynamic_control_flow()

    blocks = cfg.basic_blocks()
    assert len(blocks) == 3


def test_basic_block_fixed_schedule():
    """Test that a simple, realistic program can be scheduled as expected."""
    program = Program.parse(
        """DEFFRAME 0 "flux_tx_cz":
    TEST: 1

DEFFRAME 1 "flux_tx_iswap":
    TEST: 1

DEFFRAME 1 "flux_tx_cz":
    TEST: 1

DEFFRAME 1 "flux_tx_iswap":
    TEST: 1

DEFFRAME 2 "flux_tx_cz":
    TEST: 1

DEFFRAME 2 "flux_tx_iswap":
    TEST: 1

DEFFRAME 3 "flux_tx_cz":
    TEST: 1

DEFFRAME 3 "flux_tx_iswap":
    TEST: 1

# Simplified version
DEFCAL CZ q0 q1:
    FENCE q0 q1
    SET-PHASE q0 "flux_tx_cz" 0.0
    SET-PHASE q1 "flux_tx_iswap" 0.0
    NONBLOCKING PULSE q0 "flux_tx_cz" erf_square(duration: 6.000000000000001e-08, pad_left: 0.1e-08, pad_right: 0.1e-08)
    NONBLOCKING PULSE q1 "flux_tx_iswap" erf_square(duration: 6.000000000000001e-08, pad_left: 0.1e-08, pad_right: 0.1e-08)
    SHIFT-PHASE q0 "flux_tx_cz" 1.0
    SHIFT-PHASE q1 "flux_tx_iswap" 1.0
    FENCE q0 q1

CZ 0 1
CZ 2 3
CZ 0 2
"""
    )
    cfg = program.control_flow_graph()
    blocks = cfg.basic_blocks()
    assert len(blocks) == 1
    block = blocks[0]

    schedule = block.as_schedule_seconds(program)
    items = schedule.items

    # One for each CZ
    assert len(items) == 3

    items = {item.instruction_index: item for item in items}

    # The first CZ should start at 0
    assert items[0].time_span.start == 0.0

    # The first CZ should start and end at the same times
    assert items[0].time_span.start == items[1].time_span.start
    assert items[0].time_span.duration == items[1].time_span.duration

    # The third CZ should start when the first and second ones end and be of the same duration
    assert items[0].time_span.start + items[0].time_span.duration == items[2].time_span.start
    assert items[0].time_span.duration == items[2].time_span.duration


def test_filter_instructions(snapshot: SnapshotAssertion):
    input = """DECLARE foo REAL[1]
DEFFRAME 1 "rx":
\tHARDWARE-OBJECT: "hardware"
DEFCAL I 1:
\tDELAY 0 1
DEFGATE BAR AS MATRIX:
\t0, 1
\t1, 0

H 1
CNOT 2 3
"""
    program = Program.parse(input)
    program_without_quil_t = program.filter_instructions(lambda instruction: not instruction.is_quil_t())
    assert program_without_quil_t.to_quil() == snapshot


def test_calibration_expansion():
    """
    Assert that program calibration expansion happens as expected and that the source map is correct.
    """
    import inspect

    program_text = inspect.cleandoc(
        """
        DEFCAL X 0:
            Y 0

        DEFCAL Y 0:
            Z 0

        X 0
        Y 0
        """
    )
    program = Program.parse(program_text)

    expanded_program, source_map = program.expand_calibrations_with_source_map()

    expected_program_text = inspect.cleandoc(
        """
        DEFCAL X 0:
            Y 0

        DEFCAL Y 0:
            Z 0

        Z 0
        Z 0
        """
    )

    assert expanded_program.to_quil() == Program.parse(expected_program_text).to_quil()

    # The X at index 0 should have been replaced with a Z at index 0
    targets = source_map.list_targets_for_source_index(0)
    match targets:
        case [InstructionTarget.Calibration(expanded)]:
            pass
        case _:
            assert False, f"unexpected targets: {targets}"

    assert expanded.range == range(0, 1)
    assert source_map.list_sources_for_target_index(0) == [0]

    # The Y at index 1 should have been replaced with a Z at index 1
    targets = source_map.list_targets_for_source_index(1)
    match targets:
        case [InstructionTarget.Calibration(expanded)]:
            pass
        case _:
            assert False, f"unexpected targets: {targets}"

    assert expanded.range == range(1, 2)
    assert source_map.list_sources_for_target_index(1) == [1]

    # There is no source index 2 and so there should be no mapping
    assert source_map.list_targets_for_source_index(2) == []


def test_defgate_sequence_expansion_triple_recursive():
    """
    Assert that program sequence gate definition expansion happens as expected and that the source map
    is correct.
    """
    import inspect

    program_text = inspect.cleandoc(
        """
        DEFGATE some_u2_cycle(%param1, %param2, %param3, %param4, %param5, %param6) a b AS SEQUENCE:
            pmw3(%param1, %param2, %param3) a
            pmw3(%param4, %param5, %param6) b

        DEFGATE pmw3(%param1, %param2, %param3) a AS SEQUENCE:
            pmw(%param1) a
            pmw(%param2) a
            pmw(%param3) a

        DEFGATE pmw(%param1) a AS SEQUENCE:
            RZ(%param1) a
            RX(pi/2) a
            RZ(-%param1) a

        some_u2_cycle(-pi, -pi/2, -pi/4, pi/4, pi/2, pi) 0 1
        """
    )
    program = Program.parse(program_text)
    expanded_program, source_map = program.expand_defgate_sequences_with_source_map()

    expected_program_text = inspect.cleandoc(
        """
        RZ(-pi) 0
        RX(pi/2) 0
        RZ(-(-pi)) 0
        RZ(-pi/2) 0
        RX(pi/2) 0
        RZ(-(-pi/2)) 0
        RZ(-pi/4) 0
        RX(pi/2) 0
        RZ(-(-pi/4)) 0

        RZ(pi/4) 1
        RX(pi/2) 1
        RZ(-(pi/4)) 1
        RZ(pi/2) 1
        RX(pi/2) 1
        RZ(-(pi/2)) 1
        RZ(pi) 1
        RX(pi/2) 1
        RZ(-(pi)) 1
        """
    )

    assert expanded_program.to_quil() == Program.parse(expected_program_text).to_quil()

    # The X at index 0 should have been replaced with a Z at index 0
    targets = source_map.list_targets_for_source_index(0)
    match targets:
        case [InstructionTarget.DefGateSequence(expanded)]:
            pass
        case _:
            assert False, f"unexpected targets: {targets}"

    assert expanded.range == range(0, 18)
    assert source_map.list_sources_for_target_index(0) == [0]
    assert source_map.list_sources_for_target_index(10) == [0]
    assert source_map.list_sources_for_target_index(17) == [0]

    # There is no source index 1 and so there should be no mapping
    targets = source_map.list_targets_for_source_index(1)
    assert len(targets) == 0


def test_defgate_sequence_expansion_with_filtering():
    """
    Assert that program sequence gate definition expansion happens as expected and that the source map
    is correct.
    """
    import inspect

    program_text = inspect.cleandoc(
        """
        DECLARE ro BIT[2]

        DEFGATE seq1(%param1) a AS SEQUENCE:
            RZ(%param1) a
            seq2() a

        DEFGATE seq2() a AS SEQUENCE:
            H a

        DEFGATE seq3() a AS SEQUENCE:
            X a

        seq1(pi/2) 0
        seq2() 1
        seq3 2

        MEASURE 0 ro[0]
        MEASURE 1 ro[1]
        """
    )
    program = Program.parse(program_text)
    expansion_filter = lambda key: key != "seq1"
    expanded_program, source_map = program.expand_defgate_sequences_with_source_map(predicate=expansion_filter)

    expected_program_text = inspect.cleandoc(
        """
        DECLARE ro BIT[2]
        DEFGATE seq1(%param1) a AS SEQUENCE:
            RZ(%param1) a
            seq2() a

        DEFGATE seq2() a AS SEQUENCE:
            H a

        seq1(pi/2) 0
        H 1
        X 2
        MEASURE 0 ro[0]
        MEASURE 1 ro[1]
        """
    )

    assert expanded_program.to_quil() == Program.parse(expected_program_text).to_quil()

    targets = source_map.list_targets_for_source_index(0)
    match targets:
        case [InstructionTarget.Unmodified(0)]:
            pass
        case _:
            assert False, f"unexpected targets: {targets}"

    assert source_map.list_sources_for_target_index(0) == [0]
    assert source_map.list_sources_for_target_index(1) == [1]
    assert source_map.list_sources_for_target_index(2) == [2]

    targets = source_map.list_targets_for_source_index(1)
    match targets:
        case [InstructionTarget.DefGateSequence(expanded)]:
            pass
        case _:
            assert False, f"unexpected targets: {targets}"

    assert expanded.range == range(1, 2)

def test_defgate_sequence_expansion_with_error_handling():
    """
    Ensure that filter function errors are handled as expected during defgate sequence expansion.
    """
    def filter_fn(key: str) -> bool:
        raise KeyError(f"Gate {key} not found")
    program = Program.parse(
        inspect.cleandoc(
        """
        DEFGATE seq1() a AS SEQUENCE:
            H a

        seq1() 0
        """
        )
    )

    with pytest.raises(BaseException):
        _, _ = program.expand_defgate_sequences_with_source_map(predicate=filter_fn)



import json
from pathlib import Path
from typing import Iterable

from pyo3_linter import process_dir, find_possible_mistakes, Issue, Package, PackageKind


def package_to_dict(package: Package) -> dict:
    return {
        name: {
            "submodule": mod.submodule,
            "props": mod.props,
            "items": [
                {
                    "rust_name": item.rust_name,
                    "python_name": item.python_name,
                    "kind": item.kind.name,
                    "props": item.props,
                    "stub_attr": {
                        "kind": item.stub_attr.kind.name,
                        "module": item.stub_attr.module,
                    } if item.stub_attr else None,
                }
                for item in sorted(mod)
            ]
        }
        for name, mod in package.items()
    }


def issues_to_list(issues: Iterable[Issue]) -> list:
    return [
        {
            "package_kind": issue.package_kind.name,
            "message": issue.message,
        }
        for issue in issues
    ]


def list_to_issues(issues: list[dict]) -> set[Issue]:
    return {
        Issue(package_kind = PackageKind(issue["package_kind"]), message=issue["message"])
        for issue in issues
    }


def _save_generated(data, paths):
    for d, p in zip(data, paths):
        with open(p, "w") as f:
            json.dump(d, f, indent=2)


def test_find_mistakes(root: Path, generating: bool):
    """Processes files under the `root` and compares the results to expected output,
    stored in JSON files under that root directory.
    """

    annotated, exported = process_dir(root)
    issues = find_possible_mistakes(annotated, exported)

    # These are dicts/lists so they'll be valid JSON and easier to manipulate/check.
    got_data, expected_paths = list(zip(*(
        ( package_to_dict(annotated), (root / "expected-annotated.json") ),
        ( package_to_dict(exported), (root / "expected-exported.json") ),
        ( issues_to_list(issues), (root / "expected-issues.json") ),
    )))

    if generating:
        _save_generated(got_data, expected_paths)
        expected_data = got_data
    else:
        for p in expected_paths:
            assert p.exists() and p.is_file(), f"missing comparison file for {root.name}"
        expected_data = [ json.loads(p.read_text()) for p in expected_paths ]

    for got, expected in zip(got_data, expected_data):
        if isinstance(got, list):
            # Compare as sets, because order isn't important.
            got = list_to_issues(got)
            expected = list_to_issues(expected)
        assert got == expected




"""
A lint helper for ``quil-rs``'s PyO3 wrappers.

You'd typically use this by giving the root source path to the `process_dir` function,
then pass the resulting `Package`s to the `find_possible_mistakes` function,
getting back a list of `Issue`s that you can choose to print or further process.
You can instead (or in addition) pass the `Package`s to `print_package_info`
to get pretty-printed information about the overall package structure inferred by this code.

This works purely on the textual source, so it may make mistakes,
especially among elements that share the same name.
In particular, if there are multiple ``#[pyfunction]`` with the same Rust name,
and at least one of them is added to a module, this script won't know which one it was.
With ``pyo3_stub_gen`` annotations, the script can usually make a better guess about the intent,
but without one, it just prints all the functions it finds with a guess about the export.
"""

from dataclasses import dataclass, replace
from itertools import accumulate, chain
from pathlib import Path
from typing import TYPE_CHECKING
import logging
import re

from .reader import (
    join_lines,
    iter_delim,
    read_file,
    Line,
    Lines,
    skip,
)

from .package import (
    Item,
    Kind,
    Module,
    Package,
    PackageKind,
    PyO3Props,
    StubAttr,
    StubKind,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Issue:
    package_kind: PackageKind
    message: str


def process_dir(root: Path) -> tuple[Package, Package]:
    """Process sources recursively from the root path,
    returning the ``(annotated, exported)`` `Package`s.
    """

    annotated, exported = Package(), Package()
    for path in root.rglob("*.rs"):
        path = path.relative_to(root)
        extract_items_from_file(root, path, annotated, exported)
    guess_function_modules(annotated, exported)
    return annotated, exported


def guess_function_modules(annotated: Package, exported: Package):
    """Try to match functions to modules.

    The ``pyfunction`` annotation doesn't include a ``module`` property,
    so we can only guess the module it belongs to if we can find a matching export
    (that is, a ``pymodule`` it appears to have been added to).
    """

    builtins = annotated["builtins"]
    funcs = [f for f in builtins if f.kind is Kind.Function]
    for func in funcs:
        matching = (
            name
            for name, mod in exported.items()
            for item in mod
            if (item.kind is Kind.Function and item.rust_name == func.rust_name)
        )
        if modname := next(matching, None):
            logger.info(f"Guessing module '{modname}' for function: {func}")
            exported[modname].add(func)
            builtins.discard(func)
        else:
            logger.warning(f"No module found for function: {func}")


def find_possible_mistakes(
    annotated: Package, exported: Package, found_exports: set[str] | None = None
) -> list[Issue]:
    """Return potential mistakes between the `annotated` and `exported` ``Package``s.

    The `found_exports` parameter is a set of module names considered exported implicitly.
    If you're using this to lint code other than ``quil-rs``, you'll probably want to change them,
    but keep in mind, this linter is designed with consideration of macros used in ``quil-rs``.
    If they are not given, it uses these defaults:

    - "builtins": This is implicitly "exported".
    - "quil": This is the name of the top-level package, so it doesn't need an export.
    - "_quil": This is the name of the internal package produced by the build process,
        and its contents are exported as the top-level package within the ``__init__.py``.

    The split between the "internal" ``_quil`` and "external" ``quil`` packages
    is based on ``maturin's recommendations <https://www.maturin.rs/project_layout.html>_``,
    specifically for "Adding Python type information".
    """

    logger.debug("Checking that annotated items are added to the correct module.")
    issues: list[Issue] = []


    for method in annotated._methods:
        rust_type, _ = method.rust_name.split("::", 1)
        if (found := annotated.find_rust(rust_type)) is None:
            msg = "\n".join(
                (
                    f"Method '{method.python_name}' was not matched to a Rust type.",
                    f"   {method}",
                )
            )
            issues.append(Issue(PackageKind.Annotation, msg))
            continue

        module, item = found
        if method.python_name.startswith("Py") or method.python_name.startswith("py_"):
            msg = "\n".join(
                (
                    f"Suspicious method name for '{module}.{item.python_name}.{method.python_name}'",
                    f"     ({method})",
                    (
                        "     "
                        "hint: did you forget a '#[pyo3(name = "
                        f'"{method.python_name.lower().removeprefix("py").removeprefix("_")}"'
                        ')] attribute?'
                    ),
                )
            )
            issues.append(Issue(PackageKind.Annotation, msg))


    if found_exports is None:
        found_exports = {"builtins", "quil", "._quil"}

    for module, anno in sorted(annotated.items()):
        if module == "builtins":
            for func in (f for f in anno if f.kind is Kind.Function):
                msg = "\n".join(
                    (
                        f"Function  '{func.python_name}' does not appear to be exported",
                        f"   {func}",
                    )
                )
                issues.append(Issue(PackageKind.Annotation, msg))
            continue

        # First check if we found the module itself.
        logger.debug(f"Looking for `#[pymodule]` matching '{module}'")
        expm = exported.get(module)
        if expm is None:
            items = "\n".join(
                f"    {item} (props: {item.props}, text: {item.line})" for item in anno
            )
            msg = "\n".join(
                (
                    f"Module '{module}' does not appear to be exported",
                    f"  Module found in: {anno.path}",
                    f"  Items in module:",
                    items,
                )
            )
            issues.append(Issue(PackageKind.Annotation, msg))
            continue

        parts = module.split(".")
        found_exports.add(parts[0])
        found_exports |= set(accumulate(parts, lambda parent, sub: f"{parent}.{sub}"))

        # When parsing a ``pymodule``, we know the Rust name,
        # but if it has a different Python name, we only know when we find the ``pyclass``,
        # so here we map Rust names to items, and if we can't find an item in an export,
        # we'll try to find it by Rust name in this collection,
        # and if it has a matching `kind`, we'll assume it's the same item.
        rust_to_py_anno = {item.rust_name: item for item in anno}
        logger.debug(
            f"Checking for items added to '{module}' but not annotated as such."
        )
        for item in sorted(expm):
            if (
                item not in anno
                and (anno_item := rust_to_py_anno.get(item.rust_name))
                and anno_item.kind.is_py_like(item.kind)
            ):
                replacement = replace(item, python_name=anno_item.python_name)
                logger.debug(f"Replacing {item} with {replacement} in '{module}'.")
                expm.discard(item)
                expm.add(replacement)
                item = replacement

            if not (item in anno or item.kind is Kind.Function):
                if alt := exported.find_rust(item):
                    hint = (
                        f"\n     (hint: item may belong to module '{alt[0]}': {alt[1]})"
                    )
                else:
                    hint = ""
                msg = "\n".join(
                    (
                        f"  Wrong or missing module annotation for '{module}.{item.python_name}'",
                        f"     ({item}){hint}",
                    )
                )
                issues.append(Issue(PackageKind.Export, msg))

        logger.debug(f"Checking that items marked to belong to '{module}' are added.")
        for item in sorted(anno):
            if item.python_name.startswith("Py") or item.python_name.startswith("py_"):
                msg = "\n".join(
                    (
                        f"  Suspicious name for '{module}.{item.python_name}'",
                        f"     ({item})",
                        (
                            "     "
                            "hint: did you forget a '#[pyo3(name = "
                            f'"{item.python_name.strip("Py")})]" attribute?'
                        ),
                    )
                )
                issues.append(Issue(PackageKind.Export, msg))

            if item not in expm:
                msg = "\n".join(
                    (
                        f"  Couldn't find export for '{module}.{item.python_name}'",
                        f"     ({item})",
                    )
                )
                issues.append(Issue(PackageKind.Export, msg))

            if item.stub_attr is None:
                msg = "\n".join(
                    (
                        f"  Didn't find a stub annotation for '{module}.{item.python_name}'",
                        f"     ({item})",
                    )
                )
                issues.append(Issue(PackageKind.Export, msg))

            elif item.stub_attr.kind is StubKind.Enumeration:
                if item.props.get("rename_all") != "SCREAMING_SNAKE_CASE":
                    msg = "\n".join(
                        (
                            (
                                f"  Simple enum '{module}.{item.python_name}' doesn't have "
                                'rename_all = "SCREAMING_SNAKE_CASE"'
                            ),
                            f"     ({item})",
                        )
                    )
                    issues.append(Issue(PackageKind.Annotation, msg))

            elif item.stub_attr.kind is StubKind.ComplexEnumeration:
                if item.rust_name not in expm._fixed_enums:
                    msg = "\n".join(
                        (
                            f"  Didn't find `fix_complex_enums!` for '{module}.{item.python_name}'",
                            f"     ({item})",
                        )
                    )
                    issues.append(Issue(PackageKind.Export, msg))

    logger.debug(f"Found exports: {', '.join(found_exports)}")
    logger.debug(f"Expected exported modules: {', '.join(exported.keys())}")
    for module in exported.keys() - found_exports:
        msg = f"No annotations found for module '{module}'"
        issues.append(Issue(PackageKind.Export, msg))

    return issues


def print_package_info(annotated: Package) -> None:
    for module, anno in sorted(annotated.items()):
        items = sorted(anno)
        all_enums = [item for item in items if item.kind is Kind.Enumeration]
        complex_enums = [
            e
            for e in all_enums
            if (stubs := e.stub_attr) and stubs.kind is StubKind.ComplexEnumeration
        ]

        enums = [e for e in all_enums if e not in complex_enums]
        errors = [item for item in items if item.kind is Kind.Error]
        structs = [item for item in items if item.kind is Kind.Struct]

        to_print = [
            ("Complex Enums", complex_enums),
            ("Enums", enums),
            ("Errors", errors),
            ("Structs", structs),
        ]

        for category, items in to_print:
            if not items:
                continue

            print(f"\n---- {module} {category} ----")
            for item in items:
                full_qual = f"{module}.{item.python_name}"
                print(
                    f"{full_qual:50} | {item.rust_name:20} | {item.path}:{item.line.num}"
                )

                if category == "Structs":
                    print("  ", item.props - {"name", "module"})


def _meta(regex: str) -> str:
    r"""Create an RE to match content of an attribute; i.e., match ``#[{regex}]``."""
    return rf"#\[\s*?{regex}\s*?\]"


def _cfg(regex: str, cond: str = r"[^,]+") -> str:
    r"""Create an RE to match content of a configurable attribute;
    i.e., match either ``#[{regex}]`` or ``#[cfg_attr({cond}, {regex})]``.
    """
    return _meta(rf"(?:cfg_attr\s*?\({cond},\s*?{regex}\s*?\)|{regex})")


CONFIG_ATTR = r'(?:cfg_attr\(feature\s*?=\s*?"python",\s*?)?'
PYITEM_RE = re.compile(_cfg(rf"(?:pyo3::)?(?:pyclass|pyfunction)\s*(?:\((.*?)\))?"))
PYMODULE_RE = re.compile(_cfg(f"(?:pyo3::)?pymodule"))
PYMETHODS_RE = re.compile(_cfg(r"(?:pyo3::)?pymethods"))
PYO3_RE = re.compile(_cfg(r"pyo3\(([^)]+)\)"))
GETTER_SETTER_RE = re.compile(_cfg(r"(getter|setter)\(([^)]+)\)"))
ITEM_RE = re.compile(
    r"(?P<kind>struct|enum|fn)\s+(?:\(\s*)?(?P<rust_name>[A-Za-z_][A-Za-z0-9_]*)"
)
ADD_ERROR_RE = re.compile(
    r'\.add\(\s*"(?P<python_name>[^"]*?)"[^<]*<(?P<rust_name>[^>]*?)>'
)
ADD_CLASS_RE = re.compile(
    r"\.add_class::<(?P<rust_name>[^>]*?)>|"
    r"\.add_function\(wrap_pyfunction!\((?P<func_name>[^>]*?),.*\)"
)
STUB_GEN_RE = re.compile(
    _cfg(
        r"(?:(?:pyo3_stub_gen::)?derive::)?gen_stub_py(class(?:(?:_complex)?_enum)?|methods|function)(?:\((\s*?[^)]+)\))?"
    )
)

def _pymethods(annotated: Package, path: Path, lines: Lines):
    """Process the body of a ``#[pymethods]`` annotated ``impl`` block."""

    # Skip over additional `#[...]` lines.
    while (line := next(lines, None)) and line.text.strip().startswith("#["):
        line = join_lines(iter_delim(lines, "[]", first=line))

    if line is None:
        logger.error(f"Unexpected EOF processing {path}")
        return

    match line.text.split():
        case ["impl", rust_type, "{"] | ["impl", _, "for", rust_type, "{"]:
            logger.info(f"Processing impl block for {rust_type}.")
            pass
        case _:
            logger.error(f"Expected impl block, but found {line}.")
            return

    # Limit ourselves to the impl block.
    lines = iter_delim(lines, "{}")

    props = PyO3Props()
    while (line := next(lines, None)):
        if line.text.strip().startswith("#["):
            line = join_lines(iter_delim(lines, "[]", first=line))
            if m := PYO3_RE.search(line.text):
                props |= PyO3Props.parse(m.group(1) or m.group(2))
            elif m := GETTER_SETTER_RE.search(line.text):
                props["name"] = m.group(2) or m.group(4)

        if not (item_match := ITEM_RE.search(line.text)):
            continue

        kind: Kind = Kind(item_match.group("kind"))
        rust_name: str = item_match.group("rust_name")

        if not isinstance(rust_name, str):
            logger.error(f"Couldn't find Rust name for {path}@{line}: {kind}")
            return

        python_name = props.get("name") or props.get("getter")

        item = Item(
            kind=kind,
            python_name=props.gets("name", rust_name),
            rust_name=f"{rust_type}::{rust_name}",
            path=path,
            line=line,
        )
        assert item.kind is Kind.Function, str(item)
        if item.python_name.startswith("Py") or item.python_name.startswith("py_"):
            logger.warning(f"Suspicious name for {path}@{line}: {item}")

        annotated._methods.add(item)

        # Skip the actual function body.
        skip(lines, "{}", first=line)


def _pyitem(
    annotated: Package,
    path: Path,
    lines: Lines,
    pyclass_match: re.Match,
    last_stub: StubAttr | None = None,
) -> tuple[Module, Item] | tuple[None, None]:
    """Process #[pyclass] and #[pyfunction] annotated items,
    adding them as annotated package details.

    On success, this returns ``(module, item)``,
    where ``module`` is the `Module` where `Item` ``item`` is added.
    """

    props = PyO3Props.parse(pyclass_match.group(1) or pyclass_match.group(2) or "")

    # Skip over additional `#[...]` lines, looking for more props.
    while (line := next(lines, None)) and line.text.strip().startswith("#["):
        line = join_lines(iter_delim(lines, "[]", first=line))
        if m := PYO3_RE.search(line.text):
            props |= PyO3Props.parse(m.group(1) or m.group(2))

    if line is None:
        logger.error(f"Unexpected EOF processing {path}")
        return None, None

    if not (item_match := ITEM_RE.search(line.text)):
        logger.error(f"Failed to extract item details for {path}@{line}")
        return None, None

    kind: Kind = Kind(item_match.group("kind"))
    rust_name: str = item_match.group("rust_name")

    if not isinstance(rust_name, str):
        logger.error(f"Couldn't find Rust name for {path}@{line}: {kind}")
        return None, None

    item = Item(
        kind=kind,
        python_name=props.gets("name", rust_name),
        rust_name=rust_name,
        path=path,
        line=line,
        props=props,
        stub_attr=last_stub,
    )

    if item.python_name.startswith("Py") or item.python_name.startswith("py_"):
        logger.warning(f"Suspicious name for {path}@{line}: {item}")

    if last_stub is None:
        logger.warning(f"No stub annotation found for {path} {item}.")
    elif not last_stub.kind.is_like(item.kind):
        logger.warning(
            f"Mismatched stub {last_stub} (vs {item.kind}) found for {path} {item}."
        )
    elif last_stub.kind is StubKind.Function and last_stub.module:
        logger.info(f"Moving fn to {last_stub.module}")
        module = last_stub.module

    module = props.gets("module", "builtins")
    annotated[module].add(item)
    return annotated[module], item


def _py_source_map(annotated: Package, path: Path, line: Line, lines: Lines):
    """Process the content of the ``py_source_map!`` macro."""

    # Synthesize the match.
    pyclass_match = PYITEM_RE.search("#[pyclass]")
    assert pyclass_match is not None
    last_stub = StubAttr(StubKind.Class)

    # Read the macro content.
    body = iter_delim(chain((line,), lines), "{}")
    next(body)  # Skip the first line, which contains the opening delimiter.
    _pyitem(annotated, path, body, pyclass_match, last_stub)  # SourceMap
    _pyitem(annotated, path, body, pyclass_match, last_stub)  # SourceMapEntry


def _exceptions(annotated: Package, path: Path, line: Line, lines: Lines):
    """Process the content of the ``exception!`` and ``create_exception!`` macros."""

    body = join_lines(iter_delim(lines, "()", first=line))
    if "create_exception!(" in line:
        module, err_name, _ = body.text.split(",", maxsplit=2)
    else:
        _, module, err_name, _ = body.text.split(",", maxsplit=3)

    module = module[module.find("!") + 2 :].strip()

    item = Item(
        kind=Kind.Error,
        python_name=err_name.strip(),
        rust_name=err_name.strip(),
        path=path,
        line=line,
        stub_attr=StubAttr(kind=StubKind.Class, module=module),
    )
    annotated[module].add(item)


def _impl_instruction(exported: Package, path: Path, line: Line, lines: Lines):
    """Process the input to the ``impl_instruction!`` macro."""

    exported["quil.instructions"].update(
        Item(
            kind=Kind.Class,
            python_name=rust_name,
            rust_name=rust_name,
            path=path,
            line=line,
        )
        for name in join_lines(iter_delim(lines, "[]", first=line))
        .text.replace(" ", "")
        .removeprefix("impl_instruction!([")
        .removesuffix("]);")
        .split(",")
        if (rust_name := name.partition("[")[0].strip()) != ""
    )


def _fix_complex_enums(module: Module, line: Line, lines: Lines):
    """Process the input to the ``_fix_complex_enums!`` macro."""

    module._fixed_enums.update(
        name
        for n in (
            join_lines(iter_delim(lines, "()", first=line))
            .text.replace(" ", "")
            .removeprefix("fix_complex_enums!(")
            .removesuffix(");")
            .split(",")
        )
        if (name := n.strip()) != "" and name != "py"
    )


def _pymod(exported: Package, path: Path, lines: Lines):
    """Process the body of a #[pymodule] annotated function."""

    mod_line = -1
    props = PyO3Props()
    while (line := next(lines, None)) and line.text.strip().startswith("#["):
        line = join_lines(iter_delim(lines, "[]", first=line))
        if m := PYO3_RE.search(line.text):
            props |= PyO3Props.parse(m.group(1) or m.group(2))
            mod_line = line.num
            break

    if mod_line < 0:
        logger.error(f"Could not find module properties for {path}.")

    if line is None:
        return

    name = props.gets("name", "")
    module = f"{props.gets('module', '')}.{name}"
    if name == "_quil":
        module = "quil"
    exported[module].props |= props

    body = iter_delim(lines, "{}", first=line)
    while line := next(body, None):
        if "fix_complex_enums!(" in line.text:
            _fix_complex_enums(exported[module], line, lines)
            continue

        if "add" in line.text:
            line = join_lines(iter_delim(lines, "()", first=line))

        if add_match := ADD_ERROR_RE.search(line.text):
            python_name, rust_name = add_match.groups(("python_name", "rust_name"))
            if TYPE_CHECKING:
                assert isinstance(python_name, str)
                assert isinstance(rust_name, str)

            rust_name = rust_name[rust_name.rfind(":") + 1 :]
            item = Item(
                kind=Kind.Error,
                python_name=python_name,
                rust_name=rust_name,
                path=path,
                line=line,
            )
            exported[module].add(item)

        elif add_match := ADD_CLASS_RE.search(line.text):
            if rust_name := add_match.group("rust_name"):
                kind = Kind.Class
                python_name = rust_name
            else:
                kind = Kind.Function
                python_name = rust_name = add_match.group("func_name")

            item = Item(
                kind=kind,
                python_name=python_name,
                rust_name=rust_name,
                path=path,
                line=line,
            )
            exported[module].add(item)

        mod = exported[module]
        mod.submodule = props.is_("submodule")
        mod.path = path
        mod.line_num = mod_line


def extract_items_from_file(root: Path, path: Path, annotated: Package, exported: Package):
    """Update `annotated` and `exported` with, respectively,
    the items that are annotated to belong to a module
    and the items actually added to the module.
    """

    lines = read_file(root / path)
    last_stub: StubAttr | None = None
    while line := next(lines, None):
        # Handle macros:
        if "py_source_map!" in line:
            logger.info(f"Discovered SourceMap in {path}: {line}")
            _py_source_map(annotated, path, line, lines)
            continue

        if "impl_instruction!" in line:
            logger.info(f"Discovered impl_implementation! in {path}: {line}")
            _impl_instruction(exported, path, line, lines)
            continue

        elif re.search(r"\b(?:create_)?exception!", line.text):
            logger.info(f"Discovered Error in {path}: {line}")
            _exceptions(annotated, path, line, lines)
            continue

        # Skip non-attributes:
        if not line.text.strip().startswith("#["):
            last_stub = None
            continue

        line = join_lines(iter_delim(lines, "[]", first=line))
        logger.debug(f"Found Attribute in {path}: {line}.")

        # Note that gen_stub_* attributes must come before pyo3 attributes.
        if stubgen_match := STUB_GEN_RE.search(line.text):
            logger.info(f"Discovered stub_gen attr in {path}: {line}")
            last_stub = StubAttr.from_match(stubgen_match)
        elif "gen_stub" in line and "override_" not in line:
            logger.error(f"We're probably missing this annotation: {path} {line}")

        if pyclass_match := PYITEM_RE.search(line.text):
            logger.info(f"Discovered Item in {path}: {line}")
            _pyitem(annotated, path, lines, pyclass_match, last_stub)
        elif ("pyclass" in line.text or "pyfunction" in line.text) and not (
            "gen_stub" in line.text or "use " in line.text
        ):
            logger.error(f"We're probably missing this annotation: {path} {line}")

        if PYMETHODS_RE.search(line.text):
            logger.info(f"Discovered impl block in {path}: {line}")
            if last_stub is None:
                logger.warning(f"No stub annotation found for {path} {line}.")
            _pymethods(annotated, path, chain((line,), lines))
        elif "pymethods" in line.text and not (
            "gen_stub" in line.text or "use " in line.text
        ):
            logger.error(f"We're probably missing this annotation: {path} {line}")

        if PYMODULE_RE.search(line.text):
            logger.info(f"Discovered Module in {path}: {line}")
            _pymod(exported, path, chain((line,), lines))
        elif "pymodule" in line.text and not (
            "gen_stub" in line.text or "use " in line.text
        ):
            logger.error(f"We're probably missing this annotation: {path} {line}")



"""
These types and functions are used to read over lines of Rust source code.
"""

from collections import deque
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Iterator, TypeAlias
from typing_extensions import Self
import logging
import re

logger = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True, order=True)
class Line:
    """A line of text with the line number of the source it came from."""

    num: int
    text: str

    def __contains__(self, s: str) -> bool:
        return s in self.text

    def __iter__(self) -> Iterator:
        yield self.num
        yield self.text

    def is_blank(self) -> bool:
        return self.text.strip() == ""

    def splitat(self, index: int) -> tuple[Self, Self]:
        cls = self.__class__
        return cls(self.num, self.text[:index]), cls(self.num, self.text[index:])

    def partition(self, sep: str) -> tuple[Self, Self]:
        if (index := self.text.find(sep)) >= 0:
            return self.splitat(index)
        return self.splitat(len(self.text))


Lines: TypeAlias = Iterator[Line]


def iter_delim(
    lines: Iterator[Line],
    start="()",
    stop: str | None = None,
    /,
    first: Line | None = None,
) -> Iterator[Line]:
    """Yield lines until reading the closing delimiter.

    Lines will be yielded in the same manner as the input iterator supplies them,
    with the exception of the final line, which is split in two at the closing delimiter:
    the content before and including the closing delimiter is yielded first,
    and then any remaining content on the line is yielded after.
    Lines before and including the opening delimiter are yielded in whole,
    as this presumes you've already discovered your opening delimiter
    and are starting at that point anyway.

    If given, the initial line `first` is prepended to the `rest` of the lines,
    and generally should contain the opening delimiter.
    """
    if stop is None:
        start, stop = start

    if first is not None:
        lines = chain((first,), lines)

    reg = re.compile(rf"(?P<add>{re.escape(start)})|(?P<sub>{re.escape(stop)})")
    count = 0
    while line := next(lines, None):
        for m in reg.finditer(line.text):
            count += 1 if m.lastgroup == "add" else -1
            if count <= 0:
                yield from line.splitat(m.end())
                return
        yield line

    if count != 0 or line is not None:
        raise ValueError(
            f"mismatched delimiters: EOF when count('{start}') - count('{stop}') = {count}"
        )


def skip(
    lines: Iterator[Line],
    start: str = "()",
    stop: str | None = None,
    /,
    first: Line | None = None,
) -> Line:
    """Skip delimited text, returning only the final line, following closing delimiter."""
    return deque(iter_delim(lines, start, stop, first=first), maxlen=1).pop()


def join_lines(lines: Lines, /, first: Line | None = None, sep: str = " ") -> Line:
    """Collect lines into a single line, separated with a delimiter."""

    first = next(lines, first)
    if first is None:
        raise ValueError("no lines to iterate")
    return Line(first.num, sep.join(l.text for l in chain((first,), lines)))


MARKERS = re.compile(r"(?P<cblock>/\*)|(?P<cline>//)|(?P<macrodef>macro_rules!)")


def read_content(lines: Lines) -> Lines:
    """Iterate over lines, skipping empty lines, comments, and `macro_rules!` definitions."""

    while line := next(lines, None):
        while m := MARKERS.search(line.text):
            if m.lastgroup == "cblock":
                logger.debug(f"Skipping block comment ({line}).")
                line, rest = line.splitat(m.start())
                if not line.is_blank():
                    yield line
                line = skip(lines, "/*", "*/", first=rest)

            elif m.lastgroup == "cline":
                logger.debug(f"Skipping line comment ({line}).")
                line, _ = line.splitat(m.start())

            elif m.lastgroup == "macrodef":
                logger.debug(f"Skipping macro def ({line}).")
                line, rest = line.splitat(m.start())
                if not line.is_blank():
                    yield line
                line = skip(lines, "{}", first=rest)

        if not line.is_blank():
            yield line


def read_file(path: Path) -> Lines:
    logger.info(f"Processing {path}")
    with path.open(encoding="utf-8") as f:
        yield from read_content(
            Line(i, line.strip()) for i, line in enumerate(f, start=1)
        )
